/*
A GLUT program using OpenGL, GLUI, and GLSL.

It draws a little "particle system".

Orion Sky Lawlor, olawlor@acm.org, 2007/01/18 (Public Domain)
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include <GL/glew.h> /* GL Extensions Wrangler, http://glew.sourceforge.net/ */
#include <GL/glut.h> /* GL Utilities Toolkit, http://freeglut.sourceforge.net/ */
#include <GL/glui.h> /* GL User Interface, http://glui.sourceforge.net/ */
#include "osl/vec4.h" /* GLSL-style C++ types, by Orion Lawlor */
#include "osl/mat4.h"
using osl::mat4;

#include "osl/random.h"
#include "osl/random_vector.h"
#include "osl/random.cpp"

/* Dr. Lawlor's GLSL utilities */
#include "ogl/program.h"
#include "ogl/program.cpp"
#include "ogl/util.cpp"
#include "ogl/fast_mipmaps.c"


/* Variables updated by the GUI */
float distance=2.0; /* origin-to-camera distance */
int win_w, win_h; /**< Width and height (in pixels) of our window at last resize */

float pointsize=1.0; /* size of points */

float gravity=1.0; /* strength of gravity */

int confined=1; /* enforce boundaries? */
int reset=0; /* reset to initial conditions? */
int slowmo=0; /* slow motion by 10x ? */

/* These are rotation matrices, updated by the GLUI_Rotation spheres */
mat4 modelMat(1.0), viewMat(1.0);

/// Scale by this offset
inline void glScalefv(const float *s) {
	glScalef(s[0],s[1],s[2]);
}
/// Translate by this offset
inline void glTranslatefv(const float *s) {
	glTranslatef(s[0],s[1],s[2]);
}
/** Transpose this mat4.  If the mat4 was a rotation matrix, 
  this actually inverts the matrix. */
mat4 transpose(const mat4 &m) {
	return mat4(
		m[0][0], m[1][0], m[2][0], m[3][0],
		m[0][1], m[1][1], m[2][1], m[3][1],
		m[0][2], m[1][2], m[2][2], m[3][2],
		m[0][3], m[1][3], m[2][3], m[3][3]);
}


/******************* Particle System ************************/
class Particle {
public:
	Particle(const vec3 &pos_,const vec3 &vel_,const vec4 &color_) 
		:pos(pos_), vel(vel_), color(color_)
	{
		
	}
	
	void animate(double timestep) {
		pos+=timestep*vel;
		vec3 acc=vec3(0,-gravity,0); /* gravity accelerates particles downward */
		vel+=timestep*acc;
		
		if (confined) {
			/* Restrict motion along all three axes: confine to a box */
			double elasticity=0.8;
			int axis=1; // Y axis
			if (pos[axis]<0) {pos[axis]=0; vel[axis]*=-elasticity;}
		}
	}
	void draw(void) {
		glColor4fv(color);
		glVertex3fv(pos);
	}
	vec3 pos; /* position of particle in 3D space */
	vec3 vel; /* velocity of particle in 3D space */
	vec4 color;
};

/* A set of animating Particles */
class ParticleSystem {
public:
	ParticleSystem();
	
	void animate(double timestep) {
		for (unsigned int i=0;i<particles.size();i++)
			particles[i].animate(timestep);
	}
	void draw(void) {
		if (next>=(int)particles.size()) next=0;
		particles[next++]=makeParticle();
		glBegin(GL_POINTS);
		for (unsigned int i=0;i<particles.size();i++)
			particles[i].draw();
		glEnd();
	}
private:
	osl::Random48 r;
	std::vector<Particle> particles;
	int next;
	Particle makeParticle(void);
};

ParticleSystem::ParticleSystem()
	:r(3),next(0)
{
	for (int i=0;i<10000;i++) {
		particles.push_back(makeParticle());
	}
}

Particle ParticleSystem::makeParticle(void)
{
	/* generate a random particle location */
	vec3 l;
	while (1) { /* generate a random point */
		l=osl::nextSphere(r);
		if (l.y>0.8) break; /* that's a good point--keep it! */
	}

	/* generate a random particle color and add the particle */
	vec4 color(r.nextFloat(),r.nextFloat(),r.nextFloat(),1.0);
	return Particle(vec3(0.0),2.0*l,color);
}

ParticleSystem particles; /*<- silly global variable */

/* GLUT calls this routine every time our window needs to be redrawn */
void main_display(void) {
	static double last_elapsed=0.0;
	glFinish(); double elapsed=oglTimer();
	// Clear the color and depth buffers
	glClearColor(0.2,0.4,0.8,0.0); /* sky blue */
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); /* actually clear color and depth */
	
	// Load all needed matrices into OpenGL
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity(); /*<- clean out any old leftover matrices */
	gluPerspective(90.0, /* <- Y camera field-of-view, in degrees */
		win_w/(float)win_h, /* viewport's aspect ratio: always width over height */
		0.01, /* Near clipping plane depth */
		100.0 /* Far clipping plane depth */
	);
	glTranslatef(0,0,-distance); /* push world origin down the *unrotated* Z axis */
	glMultMatrixf(&viewMat[0][0]); /* camera orientation: could also go under GL_MODELVIEW */
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glMultMatrixf(&modelMat[0][0]);
	
	/* Set up splat texture */
	static oglTexture *splat=0;
	const int splatFrac=4; /* amount of transparent blank space to add outside splat */
	if (splat==0) {
		/* Make a new splat texture: a central dark spot
		  with fuzzy edges.  This is a greyscale texture.
		*/
		const int w=32*splatFrac, h=32*splatFrac;
		unsigned char arr[w*h];
		for (int y=0;y<h;y++)
		for (int x=0;x<w;x++) {
			double dx=x*splatFrac/(double)w-0.5;
			double dy=y*splatFrac/(double)h-0.5;
			double r=sqrt(dx*dx+dy*dy);
			double out=1.0;
			double r_min=0.05, r_max=0.5;
			if (r<r_min) out=1.0;
			else if (r>r_max) out=0.0;
			else out=1.0-(r-r_min)/(r_max-r_min);
			arr[y*w+x]=(unsigned char)(255*out);
		}
		splat=new oglTexture(arr,w,h,
			oglTexture_linear, GL_LUMINANCE,GL_UNSIGNED_BYTE,
			GL_LUMINANCE8);
	}
	
	/* Set up programmable shader. */
	static oglProgramObject *p=oglProgramFromStrings(
	"/* GLSL Vertex shader, to compute point color and size */\n"
	"uniform vec4 cameraloc; // Location of camera\n"
	"uniform float exposure; // overall exposure control--point diameter 1 unit from camera\n"
	"varying vec4 color; // Material diffuse color\n"
	"void main(void) {\n"
	"	gl_Position=gl_ModelViewProjectionMatrix*gl_Vertex;\n"
	"	color=gl_Color;\n"
	"	vec4 position=gl_ModelViewMatrix*gl_Vertex;\n"
	"	float d=distance(vec3(cameraloc),vec3(position));\n"
	"	gl_PointSize=exposure/d; // <- point diameter in pixels (drops like sqrt(1/r^2))\n"
	"	const float maxSize=1.9;\n"
	"	if (gl_PointSize<maxSize) { /* can't draw sub-pixel points */ \n"
	"		// color=vec4(0,1,0,1); /* debugging: make small points green */ \n"
	"		color.a*=(1.0/(maxSize*maxSize))*gl_PointSize*gl_PointSize; \n"
	"		color.a+=2.0/256.0; /* don't let points get *too* transparent! */\n"
	"		gl_PointSize=maxSize;\n"
	"	}\n"
	"}\n"
	,
	"/* GLSL Fragment shader, to write actual pixels */\n"
	"varying vec4 color;\n"
	"uniform sampler2D splatTex;\n"
	"void main(void) {\n"
	"	gl_FragColor=color;\n"
	"	gl_FragColor.a*=texture2D(splatTex,vec2(0.25*gl_TexCoord[0])).r;\n"
	"}\n"
	);
	p->begin();
	
	/* World location of camera: just back out all the matrix transforms */
	vec3 cameraloc=vec3(transpose(viewMat)*vec4(0,0,+distance,1));
	p->set("cameraloc",cameraloc);
	p->set("exposure",pointsize*10.0*win_h/400.0);
	splat->bind();
	p->set("splatTex",0); /* splat texture is bound to texture unit 0 */
	
	/* GL Point Sprites (magically makes texture coordinates in fragment program) */
	glEnable(GL_POINT_SPRITE_ARB);
	glTexEnvi(GL_POINT_SPRITE_ARB,GL_COORD_REPLACE_ARB,GL_TRUE);
	
	glPointSize(1.0); /* set default point size */
	/* Make gl_PointSize work in GLSL */
	glEnable(GL_VERTEX_PROGRAM_POINT_SIZE);
	
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable(GL_ALPHA_TEST); /* too agressive--loses too many points */
	glDisable(GL_DEPTH_TEST); /* don't do Z buffer (screws up overlaps, esp. w/blending) */
	glEnable(GL_BLEND);
	
	/* Draw the points */
	float anim_time=last_elapsed;
	if (slowmo) anim_time*=0.1; /* slo-motion mode */
	particles.animate(anim_time);
	if (reset) {reset=0; particles=ParticleSystem(); /* reset to initial conditions */}
	particles.draw();
	
	p->end();
	
	
	glFinish(); elapsed=oglTimer()-elapsed; if (elapsed<=1.0e-3) elapsed=1.0e-3;
	last_elapsed=elapsed;
	if (1) /* print frames per second */
	{
		/* overwrite all matrices with identity-- back to screen coordinates */
		glMatrixMode(GL_PROJECTION); glLoadIdentity();
		glMatrixMode(GL_MODELVIEW); glLoadIdentity(); 
		glColor4f(1,1,1,1);
		oglPrintf(oglVector3d(-0.9,-0.9,0.0),"%.2f ms/frame; %4.2f fps\n",
			1.0e3*elapsed,1.0/elapsed);
	}
	// Display the stuff drawn so far to the screen:
	glutSwapBuffers();
	// Repeatedly ask to be redrawn
	glutPostRedisplay();
}

/* GLUT calls this routine every time our window changes size */
void main_reshape(int w, int h) {
	win_w=w; win_h=h;
	glViewport(0, 0, (GLsizei)win_w, (GLsizei)(win_h));
}

int main(int argc, char** argv)
{
	glutInitWindowSize(800,600);
	glutInit(&argc,argv);

	// Ask for a back buffer, color buffer, and Z buffer in a window.
	int mode=GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH;
	glutInitDisplayMode (mode);
	int main_window = glutCreateWindow("GLSL Lighting Demo");

	// Turn on OpenGL extensions, like programmable shaders
	//   WARNING: you MUST call glutCreateWindow before doing this!
	//   If you don't call glewInit, or call it wrong, you'll segfault!
	glewInit();
	
	// Ask GLUT to call us when we need to redraw
	glutDisplayFunc (main_display);
	glutReshapeFunc (main_reshape);
	
	distance=2;
	/* Set up view matrix so it's tilted forward a bit */
	vec3 up=normalize(vec3(0,1,0.2));
	viewMat[1]=vec4(up,0.0);
	viewMat[2]=vec4(vec3(up.x,-up.z,up.y),0.0);

	if (1) {
		// Make some sliders (this is the only GLUI code you'll need!)
		GLUI *glui = GLUI_Master.create_glui_subwindow(main_window, 
			GLUI_SUBWINDOW_TOP);
		/* This causes the main window to be redrawn when a GLUI control changes */
		glui->set_main_gfx_window( main_window );

		(new GLUI_Spinner( glui, "Distance:", &distance))->set_float_limits(0.0, 10.0);
		(new GLUI_Spinner( glui, "Pointsize:", &pointsize))->set_float_limits(0.0,10.0);
		new GLUI_Checkbox(glui,"Reset",&reset);
		new GLUI_Column(glui);
		(new GLUI_Spinner( glui, "Gravity:", &gravity))->set_float_limits(0.0,10.0);
		new GLUI_Checkbox(glui,"Slow-mo",&slowmo);
		new GLUI_Checkbox(glui,"Confined",&confined);
		
		new GLUI_Column(glui);
		new GLUI_Rotation(glui,"Model:",&modelMat[0][0]);
		new GLUI_Column(glui,false);
		new GLUI_Rotation(glui,"View:",&viewMat[0][0]);
	}
	
	// Process events until user closes our window...
	glutMainLoop();
	
	return 0;
}
